# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import logging
from itertools import tee

import numpy as np
from six.moves import range
from six.moves import zip

logger = logging.getLogger(__name__)

BBOX_LEFT = 0
BBOX_TOP = 1
BBOX_RIGHT = 2
BBOX_BOTTOM = 3


class PKMapper(object):
    """
    Solves the following situation:

    content_pk[pk[i]] == to_map[i]

    Having two array's of the same length, with corresponding
    values:

        pk     = [1,     2,   3,   4,   5,   6]
        to_map = [1.1, 2.1, 3.1, 4.1, 5.1, 6.1]

    And an array with with (unsorted) numbers that might
    match values in pk:

        content_pk = [0, 0, 15, 3, 4, 0, 0, 1, 1, 15]


    PKMapper(pk, content_pk).apply_on(to_map, 0)

    result = [0, 0, 0, 3.1, 4.1, 0, 0, 1.1, 1.1, 0]
    """

    def __init__(self, pk, content_pk, ignore_mask=None):
        """
        :param pk: np.ndarray with pk values
        :param content_pk: np.ndarray with pk values to map to
        :param ignore_mask: boolean np.ndarray, when True the value
                            of content_pk is ignored.
        """
        # NOTE: pk should only contain unique values

        # create copies of pk an content_pk and
        # add -1 to front of pk
        # This value serves as a trash value and will be returned
        # when values in content_pk are 0 or not in pk.
        content_pk_copy = np.array(content_pk)
        pk_copy = np.concatenate((np.array([-1]), pk))

        # Apply the boolean_mask if given
        if isinstance(ignore_mask, np.ndarray):
            content_pk_copy[ignore_mask] = -1

        # Set content_pk_copy[i] = -1 if the value at i is not in pk
        content_pk_copy[np.invert(np.isin(content_pk_copy, pk))] = -1

        # sort and create an array with indices of pk
        # on content_pk
        sort_idx = np.argsort(pk_copy)
        self._indices = sort_idx[
            np.searchsorted(
                pk_copy, content_pk_copy, sorter=sort_idx)]

        # Note: self._indices[i] == 0 when the content_pk[i] could not
        # be found in pk

        del content_pk_copy
        del pk_copy

    def apply_on(self, np_array, null_value=0):
        """
        Apply the filter on np_array.

        :param np_array: the np.ndarray with values to map to content_pk
        :param null_value: the value inserted when the content_pk value
                           is zero or could not be found in pk.

        :return: an np_array with length content_pk with
                 pk[i] values replaced with np_array[i]
        """
        # Add the 'null_value' in front of np_array,
        # self._indices[i] == 0 when the content_pk[i] is not in pk
        # thus returns the null_value
        if len(np_array.shape) > 1:
            to_select_from = np.insert(np_array, 0, null_value, axis=1)
        else:
            to_select_from = np.concatenate(
                (np.array([null_value]), np_array))

        # Enable multidimensional array's.
        _filter = [slice(None)] * (
            len(np_array.shape) - 1) + [self._indices]
        return to_select_from[tuple(_filter)]


class KCUDescriptor(dict):
    """
    A dictionary-like object that provides human readable descriptions
    for threedicore kcu codes::

            0: '1d embedded line',
            1: '1d isolated line',
            2: '1d connected line',
            3: '1d long-crested structure',
            4: '1d short-crested structure',
            5: '1d double connected line',
            51: '1d2d single connected line with storage',
            52: '1d2d single connected line without storage',
            53: '1d2d double connected line with storage',
            54: '1d2d double connected line without storage',
            55: '1d2d connected line possible breach',
            56: '1d2d connected line active breach',
            57: '1d2d groundwater link',
            58: '1d2d groundwater link # diff to 57? ',
            100: '2d line',
            101: '2d obstacle (levee) line',
            150: '2d vertical link',
            -150: '2d groundwater link',
            200: '2d boundary',
            300: '2d boundary',
            400: '2d boundary',
            500: '2d boundary',

    """

    def __init__(self, *arg, **kw):
        super(KCUDescriptor, self).__init__(*arg, **kw)
        self.bound_keys_groundwater = [x for x in range(600, 1000)]
        self.bound_keys_2d = [x for x in range(200, 600)]

        self._descr = {
            0: '1d embedded line',
            1: '1d isolated line',
            2: '1d connected line',
            3: '1d long-crested structure',
            4: '1d short-crested structure',
            5: '1d double connected line',
            51: '1d2d single connected line with storage',
            52: '1d2d single connected line without storage',
            53: '1d2d double connected line with storage',
            54: '1d2d double connected line without storage',
            55: '1d2d connected line possible breach',
            56: '1d2d connected line active breach',
            57: '1d2d groundwater link',
            58: '1d2d groundwater link # diff to 57? ',
            100: '2d line',
            101: '2d obstacle (levee) line',
            150: '2d vertical link',
            -150: '2d groundwater link',
            200: '2d boundary',
            300: '2d boundary',
            400: '2d boundary',
            500: '2d boundary',
        }

    def get(self, item):
        return self.__getitem__(item)

    def values(self):
        v = list(self._descr.values())
        v.extend(['2d boundary', '2d groundwater boundary'])
        return v

    def keys(self):
        """
        Return keys
        """
        k = list(self._descr.keys())
        k += self.bound_keys_2d
        k += self.bound_keys_groundwater
        return k

    def __getitem__(self, item):
        if item in self.bound_keys_2d:
            return '2d boundary'
        elif item in self.bound_keys_groundwater:
            return '2d groundwater boundary'
        v = self._descr.get(item)
        if not v:
            raise KeyError(item)
        return v


def pairwise(iterable):
    # from https://docs.python.org/2/library/
    # itertools.html#recipes
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


def get_or_create_group(grid_file, group_name):
    """

    :param grid_file:
    :param group_name:
    :return:
    """
    gr = grid_file.get(group_name, None)
    if not gr:
        gr = grid_file.create_group(group_name)
    return gr


def _get_storage_area(storage_area):
    """

    :param storage_area: storage area entry in nodes collection.
        Either a numpy string or a bytes type
    :returns default '--' if the input is an empty string or 0., or the
        given input will be converted to a float representation
    """

    default_null = '--'
    # case python string type
    try:
        a = float(storage_area)
        if a > 0.:
            return a
    except ValueError:
        pass
    except TypeError:
        if storage_area is None:
            return default_null
        return _get_storage_area(np.asscalar(storage_area))

    return default_null



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
"""
Options
+++++++

The model _meta API is inspired by the Django _meta API. It is accessible
through the _meta attribute of a model instance, which in turn is an instance
of an threedigrid.orm.base.options.Options object.

Public methods primarily act on fields:

  - get all field instances of a model
  - get a single field instance of a model by name
  - add a single field instance to a model
  - add a field instances to a model

"""
from __future__ import unicode_literals
from __future__ import print_function

from __future__ import absolute_import
from collections import defaultdict
from collections import namedtuple
from itertools import product


import logging

from threedigrid.numpy_utils import create_np_lookup_index_for
from threedigrid.orm.base.fields import TimeSeriesCompositeArrayField
from threedigrid.orm.base.utils import _flatten_dict_values
import six

logger = logging.getLogger(__name__)


class Options(object):
    """
    class that adds meta data to a model instance. Accessible
    through the _meta API. To retrieve meta data of a field, say ``s1``:

        >>> from threedigrid.admin.gridresultadmin import GridH5ResultAdmin
        >>> f = "/code/tests/test_files/results_3di.nc"
        >>> ff = "/code/tests/test_files/gridadmin.h5"
        >>> gr = GridH5ResultAdmin(ff, f)
        >>> gr.nodes._meta.s1
        >>> s1(units=u'm', long_name=u'waterlevel',
        ...     standard_name=u'water_surface_height_above_reference_datum')

    ``s1`` is a namedtuple so you can retrieve the units attribute by the ``.``
    notation ``gr.nodes._meta.s1.units`` or using the ``_as_dict`` method
    ``gr.nodes._meta.s1._asdict()['units']``

    """
    _lookup = None  # placeholder for lookup index array

    def __init__(self, inst):
        """
        :param inst: model instance
        """
        self.inst = inst
        self._set_field_attrs()

    def get_fields(self, only_names=False):
        """
        :param only_names: omit the field instances and return a list of
            field names instead

        :returns: a dict of the models {field name: Field instance}. If
            the only_names flag is given a list of the models field
            names is returned
        """
        fields = {
            x: self.get_field(x)
            for x in self.inst._field_names if x != 'meta'
        }
        if only_names:
            return list(fields.keys())
        return fields

    def get_field(self, field_name):
        """
        :returns: the ArrayField with field_name on this instance.

        """
        return self.inst._get_field(field_name)

    def add_field(self, field_name, field, hide_private=True):
        """
        add field names to the model instance

        :param field_names: iterable of field names
        :param hide_private: fields starting with '_' will be added to the
            instance but excluded from the list of field names
        """
        if not self._source_exists(field_name):
            return
        setattr(self.inst, field_name, field)
        # remove private fields
        if hide_private and field_name.startswith('_'):
            return
        _union = set(self.inst._field_names).union({field_name})
        self.inst._field_names = _union

    def add_fields(self, fields, hide_private=True):
        """
        add fields to the models instance

        :param fields: dict of field names and fields
        :param hide_private: fields starting with '_' will be added to the
            instance but excluded from the list of field names
        """

        for field_name, field in six.iteritems(fields):
            self.add_field(field_name, field, hide_private)

    def _source_exists(self, field_name):
        """
        Checks whether field_name exists in the data source.
        Also can check composite fields

        :param field_name: name of the source field name
        :return: True if it exists False otherwise
        """
        _has_comp = hasattr(self.inst.Meta, 'composite_fields')
        _has_sub = hasattr(self.inst.Meta, 'subset_fields')
        if not _has_comp and not _has_sub:
            return field_name in list(self.inst._datasource.keys())

        sources = self.inst.Meta.composite_fields.get(field_name)

        if sources:
            return self.inst._datasource.has_any(sources)
        sources = self.inst.Meta.subset_fields.get(field_name)
        if sources:
            _sources = _flatten_dict_values(sources, as_set=True)
            return self.inst._datasource.has_any(_sources)

    def _get_meta_values(self, field_name):
        """
        fetches the values for the ``field_attrs`` entries
        of the models ``Meta`` instance
        :param field_name: name of the field
        :return: dict with field names as keys and a list with the
            corresponding meta values::

            {'s1':
                ['m', 'waterlevel',
                'water_surface_height_above_reference_datum']
            }
        """

        meta_values = defaultdict(list)
        for attr_name in self.inst.Meta.field_attrs:
            if self._is_type_composite_field(field_name):
                meta_values[field_name].append(
                    self._get_composite_meta(field_name, attr_name)
                )
            else:
                meta_values[field_name].append(
                    self.inst._datasource.attr(field_name, attr_name)
                )
        return meta_values

    def _set_field_attrs(self):
        # not all models must have a Meta instance
        if not hasattr(self.inst, 'Meta'):
            return

        # not all models with a Meta class must have a field_attrs attribute
        if not hasattr(self.inst.Meta, "field_attrs"):
            return

        # add the meta information to the instance. Uses the field name as
        # name. ``s1`` will be accessible like so for
        # example: ``gr.nodes._meta.s1``
        for _field in self.inst._field_names:
            meta_values = self._get_meta_values(_field)
            nt = namedtuple(_field, ','.join(self.inst.Meta.field_attrs))
            setattr(self, _field, nt(*meta_values[_field]))

    def _get_lookup_index(self):
        """
        creates a look up index array for the model fields which
        can be used to align result arrays to the ordering of
        grid admin arrays because it is not guaranteed  that their
        ordering is identical

        :return: numpy lookup index array
            (see ``threedigrid.numpy_utils.create_np_lookup_index_for()``
            for details) or None if the field does not have a the
            ``_needs_lookup`` attribute or if the attribute is False
        """
        if self._lookup is None:
            values = [self.inst.get_field_value(x)
                      for x in self.inst.Meta.lookup_fields]
            self._lookup = create_np_lookup_index_for(*values)

        return self._lookup

    def _get_composite_meta(self, field_name, attr_name):
        """
        get the attr entry for a composite field from the datasource

        :param field_name: field name of the model
        :param attr_name: name of the attribute from the
            ``Meta.field_attrs`` list
        :return: the attr entry from the datasource or an empty string if the
            entries in datasource are not identical for the composite fields.
            If s1 for example points to 'Mesh2D_s1' and 'Mesh1D_s1' both
            datasets must have the value 'm' for the 'units' attribute
        """
        source_names = self.inst.Meta.composite_fields.get(field_name)
        meta_attrs = [self.inst._datasource.attr(source_name, attr_name)
                      for source_name in source_names]

        if len(meta_attrs) < 2:
            return ''

        try:
            assert all(x == meta_attrs[0] for x in meta_attrs), \
                'composite fields must have the same {}. ' \
                'Failed to get meta info for field_name {} '.format(
                    attr_name, field_name)
        except AssertionError:
            return ''
        return meta_attrs[0]

    def _is_type_composite_field(self, field_name):
        """
        checks if the field is a composite field,
        like TimeSeriesCompositeArrayField
        :param field_name: name of the field
        """
        field = self.get_field(field_name)
        return isinstance(field, TimeSeriesCompositeArrayField)


class ModelMeta(type):
    """
    Metaclass for defining model meta classes.

    If you have a complexer setup you can make use of the ``ModelMeta``
    meta class that provides some field constructors behind the scenes.
    Aggregation files potentially contain a lot of different aggregation
    variables resulting in a lot of different source fields. You don't
    manually have to define all those combinations simply define a
    ``base_composition`` and a ``composition_vars`` dictionary. This computes
    all necessary composite_fields automatically.

    """

    def __new__(mcs, name, bases, namespace, **kwds):
        new_mixin = type.__new__(mcs, name, bases, dict(namespace))
        # composite_fields has been set directly, do not try to compose
        composite_fields = namespace.get('composite_fields')
        if composite_fields:
            return new_mixin

        # get base composition and vars to combine with
        base_composition = namespace.get('base_composition')
        composition_vars = namespace.get('composition_vars')
        lookup_fields = namespace.get('lookup_fields')
        base_subset_fields = namespace.get('base_subset_fields')

        # simple results
        if not composition_vars and base_composition:
            new_mixin.composite_fields = base_composition
            return new_mixin

        if composition_vars and not base_composition:
            raise AttributeError(
                'Missing base_composition attribute for the composition_vars'
            )
        # produce all possible combinations and add composite_fields
        # attribute to the class
        new_mixin.composite_fields = {}

        for k, v in six.iteritems(composition_vars):
            c_field = base_composition.get(k)
            if not c_field:
                continue
            for p in v:
                new_key = k + '_' + p
                agg_fields = ModelMeta.combine_vars(c_field, {p})
                new_mixin.composite_fields[new_key] = agg_fields

        if base_subset_fields:
            new_mixin.subset_fields = {}
            for k, v in six.iteritems(composition_vars):
                subset_dict = base_subset_fields.get(k)
                if not subset_dict:
                    continue
                s_field = list(subset_dict.values())[0]
                s_name = list(subset_dict.keys())[0]
                for p in v:
                    new_key = k + '_' + p
                    agg_fields = ModelMeta.combine_vars({s_field}, {p})
                    new_mixin.subset_fields[new_key] = {s_name: agg_fields}

        if not lookup_fields:
            return new_mixin
        lookup_field = lookup_fields[1]
        v = base_composition.get(lookup_field)
        if v:
            new_mixin.composite_fields[lookup_field] = v
        return new_mixin

    @staticmethod
    def combine_vars(prod_a, prod_b, join_str='_'):
        """Return the cartesian product of prod_a with prod_b, combined together with
        join_str.

        >>> combine_vars({'a', 'b'}, {'c', 'd'})
        ['a_c', 'a_d', 'b_c', 'b_d']

        :param prod_a: (iterable)
        :param prod_b: (iterable)
        :param join_str: (string)
        :return: (list) ∏ (set_a, set_b)
        """
        return [x[0] + join_str + x[1] for x in product(prod_a, prod_b)]



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

from __future__ import absolute_import
from threedigrid.orm.base.timeseries_mixin import ResultMixin
from threedigrid.orm.base.timeseries_mixin import AggregateResultMixin
from threedigrid.orm.base.options import ModelMeta
import six


BASE_COMPOSITE_FIELDS = {
    'au': ['Mesh2D_au', 'Mesh1D_au'],
    'u1': ['Mesh2D_u1', 'Mesh1D_u1'],
    'q': ['Mesh2D_q', 'Mesh1D_q'],
    'qp': ['Mesh2D_qp'],
    'up1': ['Mesh2D_up1'],
    '_mesh_id': ['Mesh2DLine_id', 'Mesh1DLine_id'],  # private
}

BASE_SUBSET_FIELDS = {
    'qp': {'2d_all': 'Mesh2D_qp'},
    'up1':
        {'2d_all': 'Mesh2D_up1'},
    'breach_depth':
        {'1d_all': 'Mesh1D_breach_depth'},
    'breach_width':
        {'1d_all': 'Mesh1D_breach_width'}
}


class LinesResultsMixin(ResultMixin):

    class Meta:

        # attributes for the given fields
        field_attrs = ['units', 'long_name', 'standard_name']

        composite_fields = BASE_COMPOSITE_FIELDS
        subset_fields = BASE_SUBSET_FIELDS

        lookup_fields = ('id', '_mesh_id')

    def __init__(self, **kwargs):
        """Instantiate a line with netcdf results.

        Variables stored in the netcdf and related to lines are dynamically
        added as attributes as TimeSeriesArrayField.

        :param netcdf_keys: list of netcdf variables
        :param kwargs:
        """
        super(LinesResultsMixin, self).__init__(**kwargs)


class LinesAggregateResultsMixin(AggregateResultMixin):

    class Meta(six.with_metaclass(ModelMeta)):
        field_attrs = ['units', 'long_name']

        base_composition = BASE_COMPOSITE_FIELDS
        base_subset_fields = BASE_SUBSET_FIELDS

        composition_vars = {
            'au': ['min', 'max', 'avg'],
            'u1': ['min', 'max', 'avg'],
            'q': ['min', 'max', 'avg', 'cum', 'cum_positive', 'cum_negative'],
            'up1': ['min', 'max', 'avg'],
            'qp': ['min', 'max', 'avg', 'cum', 'cum_positive', 'cum_negative'],
        }

        lookup_fields = ('id', '_mesh_id')

    def __init__(self, **kwargs):
        """Instantiate a line with netcdf results.

        Variables stored in the netcdf and related to lines are dynamically
        added as attributes as TimeSeriesArrayField.

        :param netcdf_keys: list of netcdf variables
        :param kwargs:
        """
        super(LinesAggregateResultsMixin, self).__init__(**kwargs)



from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import os
import sys
import unittest

import h5py

if sys.version_info >= (3, 3):  # noqa
    from unittest import mock
else:
    import mock
import numpy as np

from threedigrid.admin.idmapper import IdMapper
from threedigrid.admin.h5py_datasource import H5pyGroup
from threedigrid.admin.constants import TYPE_CODE_MAP
from .conftest import simple_id_map, NODE_LENGTH

test_file_dir = os.path.join(os.path.dirname(__file__), "test_files")

# the testfile is a copy of the v2_bergermeer gridadmin file
grid_file = os.path.join(test_file_dir, "gridadmin.h5")


class TestIdMapperByCode(unittest.TestCase):
    def setUp(self):
        h5py_file = h5py.File(grid_file, "r")
        self.id_mapper = IdMapper(H5pyGroup(h5py_file, "mappings")["id_map"])

    def test_get_channels_by_code(self):
        tc = TYPE_CODE_MAP["v2_channel"]
        qs = self.id_mapper.get_by_code(tc)
        assert np.all(qs["obj_code"] == tc)

    def test_get_culvert_by_code(self):
        tc = TYPE_CODE_MAP["v2_culvert"]
        qs = self.id_mapper.get_by_code(tc)
        assert np.all(qs["obj_code"] == tc)


class TestIdMapperByName(unittest.TestCase):
    def setUp(self):
        h5py_file = h5py.File(grid_file, "r")
        self.id_mapper = IdMapper(H5pyGroup(h5py_file, "mappings")["id_map"])

    def test_get_channels_by_code(self):
        tc = TYPE_CODE_MAP["v2_channel"]
        qs = self.id_mapper.get_by_name("v2_channel")
        assert np.all(qs["obj_code"] == tc)

    def test_get_culvert_by_code(self):
        tc = TYPE_CODE_MAP["v2_culvert"]
        qs = self.id_mapper.get_by_name("v2_culvert")
        assert np.all(qs["obj_code"] == tc)


@mock.patch("threedigrid.admin.idmapper.get_id_map")
def test_prepare_mapper(mocked_id_map, h5py_file, threedi_datasource):
    mocked_id_map.return_value = simple_id_map()
    IdMapper.prepare_mapper(h5py_file, threedi_datasource)
    # now a mapping should exist on the h5py file
    mapping = h5py_file["mappings"]["id_map"]
    assert mapping
    assert mapping.dtype.names == ("obj_code", "pk", "seq_id")
    assert mapping[:].size == len(TYPE_CODE_MAP.values()) * NODE_LENGTH


@mock.patch("threedigrid.admin.idmapper.get_id_map")
def test_init_idMapper(mocked_id_map, h5py_file, threedi_datasource):
    mocked_id_map.return_value = simple_id_map()
    IdMapper.prepare_mapper(h5py_file, threedi_datasource)
    mapper = IdMapper(h5py_file["mappings"]["id_map"])
    assert mapper.get_by_code(TYPE_CODE_MAP["v2_channel"]).size == NODE_LENGTH
    assert mapper.get_by_name("v2_channel").size == NODE_LENGTH



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function


from __future__ import absolute_import
from threedigrid.orm.base.filters import BaseFilter
from threedigrid.orm.base.filters import FILTER_MAP as BASE_FILTER_MAP


class GeomFilter(BaseFilter):
    def to_dict(self):
        return {
            'filter': {
                "{0}__{1}".format(
                    self.key, self.func_name): self.values
            }
        }


class BboxFilter(GeomFilter):
    """
    Bbox (geom) filter
    """
    include_intersections = False
    func_name = 'in_bbox'

    def __init__(self, key, field, values, filter_as=False):
        self.field = field
        self.key = key
        self.bbox = values
        self.values = values
        self._filter_as = filter_as
        assert hasattr(self.field, 'get_mask_by_bbox')

    def get_field_name(self):
        return self.key

    def filter(self, nparray_dict):
        return self.field.get_mask_by_bbox(
            self.bbox, nparray_dict[self.key][:],
            self.include_intersections)

    def filter_dict(self, nparray_dict, model_instance):
        """
        Filter the values in nparray_dict dictionairy by
        the filter defined in self.filter()

        :param nparray_dict:  dictionairy of np_array's.
        :param target_epsg_code: epsg_code of instance (ignored for bbox)
        """
        # Get the filter
        base_filter = self.filter(nparray_dict)
        self._do_filter(base_filter, nparray_dict)


class PointFilter(GeomFilter):
    func_name = 'contains_point'

    def __init__(self, key, field, values, filter_as=False):
        self.field = field
        self.key = key
        self.x, self.y = values
        self.values = values
        self._filter_as = filter_as
        assert hasattr(self.field, 'get_mask_by_point')

    def get_field_name(self):
        return self.key

    def filter(self, nparray_dict):
        return self.field.get_mask_by_point(
            (self.x, self.y), nparray_dict[self.key][:]
        )

    def filter_dict(self, nparray_dict, model_instance):
        """
        Filter the values in nparray_dict dictionairy by
        the filter defined in self.filter()

        :param nparray_dict:  dictionairy of np_array's.
        :param target_epsg_code: epsg_code of instance
        """
        # Get the filter
        base_filter = self.filter(nparray_dict)
        self._do_filter(base_filter, nparray_dict)


class TileFilter(GeomFilter):
    """
    Tile (geom) filter
    """
    include_intersections = False
    func_name = 'in_tile'

    def __init__(self, key, field, values, filter_as=False):
        self.field = field
        self.key = key
        self.x, self.y, self.z = values
        self.values = values
        self._filter_as = filter_as
        assert hasattr(self.field, 'get_mask_by_tile')

    def get_field_name(self):
        return self.key

    def filter(self, nparray_dict, target_epsg_code):
        return self.field.get_mask_by_tile(
            (int(self.x), int(self.y), int(self.z)),
            target_epsg_code, nparray_dict[self.key][:],
            self.include_intersections)

    def filter_dict(self, nparray_dict, model_instance):
        """
        Filter the values in nparray_dict dictionairy by
        the filter defined in self.filter()

        :param nparray_dict:  dictionairy of np_array's.
        :param target_epsg_code: epsg_code of instance
        """
        # Get the filter
        base_filter = self.filter(nparray_dict, model_instance.epsg_code)
        self._do_filter(base_filter, nparray_dict)


class BboxIntersectsFilter(BboxFilter):
    include_intersections = True
    func_name = 'intersects_bbox'


class TileIntersectsFilter(TileFilter):
    include_intersections = True
    func_name = 'intersects_tile'


class GeometryIntersectionFilter(GeomFilter):
    func_name = 'intersects_geometry'

    def __init__(self, key, field, values, filter_as=False):
        self.field = field
        self.key = key
        self.geometry = values
        self.values = values
        # geom is Shapely.geometry
        self._filter_as = filter_as
        assert hasattr(self.field, 'get_mask_by_geometry')

    def get_field_name(self):
        return self.key

    def filter(self, nparray_dict):
        return self.field.get_mask_by_geometry(
            self.geometry, nparray_dict[self.key][:],
        )

    def filter_dict(self, nparray_dict, model_instance):
        """
        Filter the values in nparray_dict dictionairy by
        the filter defined in self.filter()

        :param nparray_dict: dictionairy of np_array's.
        :param model_instance: epsg_code of instance (ignored)
        """
        # Get the filter
        base_filter = self.filter(nparray_dict)
        self._do_filter(base_filter, nparray_dict)


FILTER_MAP = dict(BASE_FILTER_MAP, ** {
    'in_tile': TileFilter,
    'in_bbox': BboxFilter,
    'contains_point': PointFilter,
    'intersects_tile': TileIntersectsFilter,
    'intersects_bbox': BboxIntersectsFilter,
    'intersects_geometry': GeometryIntersectionFilter}
)



from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import os
import unittest
import tempfile
import shutil

import numpy as np
from osgeo import ogr

from threedigrid.admin.gridadmin import GridH5Admin
from threedigrid.admin.nodes.exporters import CellsOgrExporter
from threedigrid.admin.nodes.exporters import NodesOgrExporter
from threedigrid.admin.lines.exporters import LinesOgrExporter
from threedigrid.admin.breaches.exporters import BreachesOgrExporter
from threedigrid.admin.constants import SUBSET_1D_ALL
from threedigrid.admin.constants import SUBSET_2D_OPEN_WATER
from threedigrid.admin.constants import NO_DATA_VALUE
from six.moves import range

test_file_dir = os.path.join(os.getcwd(), "tests/test_files")

# the testfile is a copy of the v2_bergermeer gridadmin file
grid_admin_h5_file = os.path.join(test_file_dir, "gridadmin.h5")


class GridAdminTest(unittest.TestCase):
    def setUp(self):
        self.parser = GridH5Admin(grid_admin_h5_file)

    def test_get_from_meta(self):
        self.assertIsNotNone(self.parser.get_from_meta("n2dtot"))
        self.assertIsNone(self.parser.get_from_meta("doesnotexist"))

    def test_get_extent_subset_onedee(self):
        extent_1D = self.parser.get_extent_subset(subset_name=SUBSET_1D_ALL)
        # should contain values
        self.assertTrue(np.any(extent_1D != NO_DATA_VALUE))

    def test_get_extent_subset_reproject(self):
        extent_1D = self.parser.get_extent_subset(subset_name=SUBSET_1D_ALL)
        extent_1D_proj = self.parser.get_extent_subset(
            subset_name=SUBSET_1D_ALL, target_epsg_code="4326"
        )
        self.assertTrue(np.all(extent_1D != extent_1D_proj))

    def test_get_extent_subset_twodee(self):
        extent_2D = self.parser.get_extent_subset(
            subset_name=SUBSET_2D_OPEN_WATER
        )
        # should contain values
        self.assertTrue(np.any(extent_2D != NO_DATA_VALUE))

    def test_get_extent_subset_combi(self):
        extent_1D = self.parser.get_extent_subset(subset_name=SUBSET_1D_ALL)
        extent_2D = self.parser.get_extent_subset(
            subset_name=SUBSET_2D_OPEN_WATER
        )
        # should be different
        self.assertTrue(np.any(np.not_equal(extent_1D, extent_2D)))

    def test_get_model_extent(self):
        model_extent = self.parser.get_model_extent()
        np.testing.assert_almost_equal(
            model_extent,
            np.array([105427.6, 511727.0515702, 115887.0, 523463.3268483]),
        )

    def test_get_model_extent_extra_extent(self):
        onedee_extra = np.array([100000.0, 90000.0, 550000.0, 580000.0])

        extra_extent = {"extra_extent": [onedee_extra]}
        model_extent = self.parser.get_model_extent(**extra_extent)
        np.testing.assert_equal(
            model_extent, np.array([100000.0, 90000.0, 550000.0, 580000.0])
        )

    def test_get_model_extent_extra_extent2(self):
        onedee_extra = np.array([106666.6, 106666.6, 550000.0, 580000.0])

        extra_extent = {"extra_extent": [onedee_extra]}
        model_extent = self.parser.get_model_extent(**extra_extent)
        np.testing.assert_almost_equal(
            model_extent, np.array([105427.6, 106666.6, 550000.0, 580000.0])
        )

    def test_properties(self):
        self.assertTrue(hasattr(self.parser, "has_groundwater"))
        self.assertTrue(hasattr(self.parser, "has_levees"))
        self.assertTrue(hasattr(self.parser, "threedicore_version"))
        self.assertTrue(hasattr(self.parser, "has_1d"))
        self.assertTrue(hasattr(self.parser, "has_2d"))
        self.assertTrue(hasattr(self.parser, "epsg_code"))
        self.assertTrue(hasattr(self.parser, "revision_hash"))
        self.assertTrue(hasattr(self.parser, "revision_nr"))
        self.assertTrue(hasattr(self.parser, "model_slug"))


class GridAdminLinesTest(unittest.TestCase):
    def setUp(self):
        self.parser = GridH5Admin(grid_admin_h5_file)
        d = tempfile.mkdtemp()
        self.f = os.path.join(d, "test_lines.shp")

    def tearDown(self):
        shutil.rmtree(os.path.dirname(self.f))

    def test_fields(self):
        # Check dtype names
        assert set(self.parser.lines._meta.get_fields().keys()) == {
            "content_pk",
            "content_type",
            "id",
            "kcu",
            "lik",
            "line",
            "line_coords",
            "line_geometries",
            "zoom_category",
        }

    def test_exporters(self):
        self.assertEqual(len(self.parser.lines._exporters), 1)
        self.assertIsInstance(
            self.parser.lines._exporters[0], LinesOgrExporter
        )

    def test_export_to_shape(self):
        self.parser.lines.to_shape(self.f)
        self.assertTrue(os.path.exists, self.f)
        s = ogr.Open(self.f)
        lyr = s.GetLayer()
        self.assertEqual(lyr.GetFeatureCount(), self.parser.lines.id.size)


class GridAdminGridTest(unittest.TestCase):
    def setUp(self):
        self.parser = GridH5Admin(grid_admin_h5_file)

    def test_fields(self):
        assert set(self.parser.grid._meta.get_fields().keys()) == {
            "id",
            "nodk",
            "nodm",
            "nodn",
            "ip",
            "jp"
        }

    @unittest.skip("TODO")
    def test_get_pixel_map(self):
        self.parser.grid.get_pixel_map()


class GridAdminNodeTest(unittest.TestCase):
    def setUp(self):
        self.parser = GridH5Admin(grid_admin_h5_file)
        d = tempfile.mkdtemp()
        self.f = os.path.join(d, "test_nodes.shp")

    def tearDown(self):
        shutil.rmtree(os.path.dirname(self.f))

    def test_fields(self):
        # Check fields
        assert set(self.parser.nodes._meta.get_fields(only_names=True)) == {
            "cell_coords",
            "content_pk",
            "coordinates",
            "id",
            "node_type",
            "seq_id",
            "zoom_category",
            "is_manhole",
            "sumax",
        }

    def test_locations_2d(self):
        self.assertGreater(len(self.parser.nodes.locations_2d), 0)
        frst = self.parser.nodes.locations_2d[0]
        # should contain three elements
        self.assertEqual(len(frst), 3)

    def test_exporters(self):
        self.assertEqual(len(self.parser.nodes._exporters), 1)
        self.assertIsInstance(
            self.parser.nodes._exporters[0], NodesOgrExporter
        )

    def test_export_to_shape(self):
        self.parser.nodes.to_shape(self.f)
        self.assertTrue(os.path.exists, self.f)
        s = ogr.Open(self.f)
        lyr = s.GetLayer()
        self.assertEqual(lyr.GetFeatureCount(), self.parser.nodes.id.size)


class GridAdminBreachTest(unittest.TestCase):
    def setUp(self):
        self.parser = GridH5Admin(grid_admin_h5_file)
        d = tempfile.mkdtemp()
        self.f = os.path.join(d, "test_breaches.shp")

    def tearDown(self):
        shutil.rmtree(os.path.dirname(self.f))

    def test_fields(self):
        # Check fields
        assert set(self.parser.breaches._meta.get_fields().keys()) == {
            "content_pk",
            "coordinates",
            "id",
            "kcu",
            "levbr",
            "levl",
            "levmat",
            "seq_ids",
        }

    def test_exporters(self):
        self.assertEqual(len(self.parser.breaches._exporters), 1)
        self.assertIsInstance(
            self.parser.breaches._exporters[0], BreachesOgrExporter
        )

    def test_export_to_shape(self):
        self.parser.breaches.to_shape(self.f)
        self.assertTrue(os.path.exists, self.f)
        s = ogr.Open(self.f)
        lyr = s.GetLayer()
        self.assertEqual(lyr.GetFeatureCount(), self.parser.breaches.id.size)


class GridAdminCellsTest(unittest.TestCase):
    def setUp(self):
        self.parser = GridH5Admin(grid_admin_h5_file)
        d = tempfile.mkdtemp()
        self.f = os.path.join(d, "test_cells.shp")

    def tearDown(self):
        shutil.rmtree(os.path.dirname(self.f))

    def test_fields(self):
        # should have also z_coordinate
        assert set(self.parser.cells._meta.get_fields().keys()) == {
            "cell_coords",
            "content_pk",
            "coordinates",
            "id",
            "node_type",
            "seq_id",
            "z_coordinate",
            "zoom_category",
            "is_manhole",
            "sumax",
            "pixel_width",
            "pixel_coords",
        }

    def test_get_id_from_xy(self):
        # should yield tow ids, one for 2d, one for groundwater
        self.assertListEqual(self.parser.cells.get_id_from_xy(1.0, 2.0), [])
        # first coordinate pair + some offset
        x = self.parser.cells.coordinates[0][1] + 0.5
        y = self.parser.cells.coordinates[1][1] + 0.5
        self.assertListEqual(self.parser.cells.get_id_from_xy(x, y), [1, 5375])

    def test_get_id_from_xy_2d_open_water(self):

        self.assertListEqual(
            self.parser.cells.get_id_from_xy(
                1.0, 2.0, subset_name="2d_open_water"
            ),
            [],
        )
        # first coordinate pair + some offset
        x = self.parser.cells.coordinates[0][1] + 0.5
        y = self.parser.cells.coordinates[1][1] + 0.5
        self.assertEqual(
            self.parser.cells.get_id_from_xy(
                x, y, subset_name="2d_open_water"
            ),
            [1],
        )

    def test_get_id_from_xy_groundwater(self):

        self.assertListEqual(
            self.parser.cells.get_id_from_xy(
                1.0, 2.0, subset_name="groundwater_all"
            ),
            [],
        )
        # first coordinate pair + some offset
        x = self.parser.cells.coordinates[0][1] + 0.5
        y = self.parser.cells.coordinates[1][1] + 0.5
        self.assertEqual(
            self.parser.cells.get_id_from_xy(
                x, y, subset_name="groundwater_all"
            ),
            [5375],
        )

    def test_exporters(self):
        self.assertEqual(len(self.parser.cells._exporters), 1)
        self.assertIsInstance(
            self.parser.cells._exporters[0], CellsOgrExporter
        )

    def test_export_to_shape(self):
        self.parser.cells.to_shape(self.f)
        self.assertTrue(os.path.exists, self.f)


class NodeFilterTests(unittest.TestCase):
    def setUp(self):
        self.parser = GridH5Admin(grid_admin_h5_file)

    def test_nodes_filter_id_eq(self):
        filtered = self.parser.nodes.filter(id=3).data["coordinates"]
        trues, = np.where(self.parser.nodes.data["id"] == 3)
        expected = self.parser.nodes.data["coordinates"][:, trues]
        self.assertTrue((filtered == expected).all())

    def test_nodes_filter_id_ne(self):
        filtered = self.parser.nodes.filter(id__ne=3).data["coordinates"]
        trues, = np.where(self.parser.nodes.data["id"] != 3)
        expected = self.parser.nodes.data["coordinates"][:, trues]
        self.assertTrue((filtered == expected).all())

    def test_nodes_filter_id_gt(self):
        filtered = self.parser.nodes.filter(id__gt=3).data["coordinates"]
        trues, = np.where(self.parser.nodes.data["id"] > 3)
        expected = self.parser.nodes.data["coordinates"][:, trues]
        self.assertTrue((filtered == expected).all())

    def test_nodes_filter_id_gte(self):
        filtered = self.parser.nodes.filter(id__gte=3).data["coordinates"]
        trues, = np.where(self.parser.nodes.data["id"] >= 3)
        expected = self.parser.nodes.data["coordinates"][:, trues]
        self.assertTrue((filtered == expected).all())

    def test_nodes_filter_id_lt(self):
        filtered = self.parser.nodes.filter(id__lt=3).data["coordinates"]
        trues, = np.where(self.parser.nodes.data["id"] < 3)
        expected = self.parser.nodes.data["coordinates"][:, trues]
        self.assertTrue((filtered == expected).all())

    def test_nodes_filter_id_lte(self):
        filtered = self.parser.nodes.filter(id__lte=3).data["coordinates"]
        trues, = np.where(self.parser.nodes.data["id"] <= 3)
        expected = self.parser.nodes.data["coordinates"][:, trues]
        self.assertTrue((filtered == expected).all())

    def test_nodes_filter_id_in(self):
        """Verify that 'in' filter returns the correct data."""
        filtered = self.parser.nodes.filter(id__in=list(range(3, 7))).data[
            "coordinates"
        ]
        trues, = np.where(
            (self.parser.nodes.data["id"] >= 3)
            & (self.parser.nodes.data["id"] < 7)
        )
        expected = self.parser.nodes.data["coordinates"][:, trues]
        self.assertTrue((filtered == expected).all())

    # TODO: fix this
    @unittest.skip(
        "This test should succeed, but in our code the reprojected tile "
        "bounds go out of bounds, which causes errors. TODO: fix this."
    )
    def test_nodes_filter_id_in_tile(self):
        # at z=0 we have a single base tile
        filtered = self.parser.nodes.filter(
            coordinates__in_tile=[0, 0, 0]
        ).data["id"]
        expected = self.parser.nodes.data["id"]
        self.assertTrue(len(filtered) != 0)
        self.assertTrue((filtered == expected).all())

    # some special cases

    def test_nodes_filter_id_eq_chained(self):
        """Eq filter can be chained."""
        filtered = (
            self.parser.nodes.filter(id=3).filter(id=3).data["coordinates"]
        )
        trues, = np.where(self.parser.nodes.data["id"] == 3)
        expected = self.parser.nodes.data["coordinates"][:, trues]
        self.assertTrue((filtered == expected).all())

    def test_nodes_chained_same_filter_id_in(self):
        """In filters can be chained."""
        filtered = (
            self.parser.nodes.filter(id__in=list(range(1, 10)))
            .filter(id__in=list(range(1, 10)))
            .filter(id__in=list(range(1, 10)))
            .data["coordinates"]
        )
        expected = self.parser.nodes.filter(id__in=list(range(1, 10))).data[
            "coordinates"
        ]
        self.assertTrue((filtered == expected).all())

    def test_nodes_chained_filter_id_in(self):
        filtered = (
            self.parser.nodes.filter(id__in=list(range(1, 8)))
            .filter(id__in=list(range(3, 7)))
            .data["coordinates"]
        )
        expected = self.parser.nodes.filter(id__in=list(range(3, 7))).data[
            "coordinates"
        ]
        self.assertTrue((filtered == expected).all())

    def test_nodes_chained_filter_id_in_2(self):
        filtered = (
            self.parser.nodes.filter(id__in=list(range(1, 10)))
            .filter(id__in=list(range(5, 20)))
            .data["coordinates"]
        )
        expected = self.parser.nodes.filter(id__in=list(range(5, 10))).data[
            "coordinates"
        ]
        self.assertTrue((filtered == expected).all())

    def test_manhole_filter(self):
        non_manholes1 = self.parser.nodes.filter(is_manhole=False).count
        non_manholes2 = self.parser.nodes.filter(is_manhole=0).count
        non_manholes3 = self.parser.nodes.filter(is_manhole__eq=0).count
        non_manholes4 = self.parser.nodes.filter(is_manhole__eq=False).count
        non_manholes5 = self.parser.nodes.filter(is_manhole__ne=1).count
        non_manholes6 = self.parser.nodes.filter(is_manhole__ne=True).count
        self.assertEqual(non_manholes1, non_manholes2)
        self.assertEqual(non_manholes2, non_manholes3)
        self.assertEqual(non_manholes3, non_manholes4)
        self.assertEqual(non_manholes4, non_manholes5)
        self.assertEqual(non_manholes5, non_manholes6)

        manholes1 = self.parser.nodes.filter(is_manhole=True).count
        manholes2 = self.parser.nodes.filter(is_manhole=1).count
        manholes3 = self.parser.nodes.filter(is_manhole__eq=1).count
        manholes4 = self.parser.nodes.filter(is_manhole__eq=True).count
        manholes5 = self.parser.nodes.filter(is_manhole__ne=0).count
        manholes6 = self.parser.nodes.filter(is_manhole__ne=False).count
        self.assertEqual(manholes1, manholes2)
        self.assertEqual(manholes2, manholes3)
        self.assertEqual(manholes3, manholes4)
        self.assertEqual(manholes4, manholes5)
        self.assertEqual(manholes5, manholes6)

        self.assertEqual(manholes1 + non_manholes2, self.parser.nodes.count)

    def test_node_filter_keeps_has_1d(self):
        """Property has_1d doesn't disappear"""
        self.assertEqual(
            self.parser.nodes.has_1d, self.parser.nodes.filter(id=1).has_1d
        )


class LineFilterTests(unittest.TestCase):
    def setUp(self):
        self.parser = GridH5Admin(grid_admin_h5_file)

    def test_lines_filter_id_eq(self):
        # from nose.tools import set_trace; set_trace()
        filtered = self.parser.lines.filter(id=3).data["line_coords"]
        trues, = np.where(self.parser.lines.data["id"] == 3)
        expected = self.parser.lines.data["line_coords"][:, trues]
        self.assertTrue((filtered == expected).all())

    def test_lines_filter_id_in(self):
        filtered = self.parser.lines.filter(id__in=list(range(3, 7))).data[
            "line_coords"
        ]
        trues, = np.where(
            (self.parser.lines.data["id"] >= 3)
            & (self.parser.lines.data["id"] < 7)
        )
        expected = self.parser.lines.data["line_coords"][:, trues]
        self.assertTrue((filtered == expected).all())



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

from __future__ import absolute_import
import json

from threedigrid.admin.utils import _get_storage_area
from threedigrid.admin.nodes.models import AddedCalculationNodes
from threedigrid.admin.nodes.models import ConnectionNodes
from threedigrid.admin.nodes.models import Manholes
from threedigrid.admin import constants
from threedigrid.geo_utils import raise_import_exception
from threedigrid.orm.base.encoder import NumpyEncoder
from six.moves import range

try:
    import geojson
except ImportError:
    geojson = None


class ManholesGeoJsonSerializer():
    def __init__(self, manholes, indent=None):
        if geojson is None:
            raise_import_exception('geojson')
        assert isinstance(manholes, Manholes)
        self._manholes = manholes
        self._indent = indent

    @property
    def geos(self):
        selection = self._manholes.to_dict()

        geos = []
        if 'coordinates' not in selection:
            raise ValueError(
                "Can't return data as geojson "
                "for selection without geometries")

        for i in range(selection['id'].shape[-1]):
            pt = geojson.Point(
                [round(x, constants.LONLAT_DIGITS)
                 for x in selection['coordinates'][:, i]])

            area = _get_storage_area(selection['storage_area'][i])

            cn_meta = [
                    ['object_type', constants.TYPE_V2_MANHOLE],
                    ['display_name', selection['display_name'][i]],
                    ['manhole_type', constants.MANHOLE_TYPES.get(
                        selection['manhole_indicator'][i])],
                    ['calculation_type', constants.CALCULATION_TYPES.get(
                        selection['calculation_type'][i])],
                    ['shape', selection['shape'][i]],
                    ['area', area],
                    ['bottom_level', "{0} [m MSL]".format(
                        selection['bottom_level'][i])],
                    ['width', "{0} [m]".format(selection['width'][i])],
                    ['drain_level', "{0} [m]".format(
                        selection['drain_level'][i])],
                    ['surface_level', "{0} [m]".format(
                        selection['surface_level'][i])],
                    ['nod idx', int(selection['id'][i])],
                    ]
            cn_props = dict(
                object_type=constants.TYPE_V2_CONNECTION_NODES,
                zoom_category=selection['zoom_category'][i],
                meta=cn_meta,
                node_idx=int(selection['id'][i])
                )
            feat = geojson.Feature(
                geometry=pt,
                properties=cn_props,
                )
            geos.append(feat)

        return geos

    @property
    def data(self):
        geos = self.geos
        return json.dumps({
            'type': 'FeatureCollection',
            'features': geos,
        }, indent=self._indent, cls=NumpyEncoder)


class ConnectionNodesGeoJsonSerializer():
    def __init__(self, connection_nodes=None, data=None, indent=None):
        if geojson is None:
            raise_import_exception('geojson')
        if connection_nodes:
            assert isinstance(connection_nodes, ConnectionNodes)
        self._data = data
        self._connection_nodes = connection_nodes
        self._indent = indent

    @property
    def geos(self):
        if self._connection_nodes:
            selection = self._connection_nodes.to_dict()
        else:
            selection = self._data

        geos = []
        if 'coordinates' not in selection:
            raise ValueError(
                "Can't return data as geojson "
                "for selection without geometries")

        for i in range(selection['id'].shape[-1]):
            pt = geojson.Point(
                [round(x, constants.LONLAT_DIGITS)
                 for x in selection['coordinates'][:, i]])
            area = _get_storage_area(selection['storage_area'][i])

            cn_meta = [
                    ['object_type', constants.TYPE_V2_CONNECTION_NODES],
                    ['storage area', area],
                    ['initial_waterlevel', "{0} [m MSL]".format(
                        selection['initial_waterlevel'][i])],
                    ['nod idx', int(selection['id'][i])],
                    ]
            cn_props = dict(
                object_type=constants.TYPE_V2_CONNECTION_NODES,
                meta=cn_meta,
                node_idx=int(selection['id'][i])
                )
            feat = geojson.Feature(
                geometry=pt,
                properties=cn_props,
                )
            geos.append(feat)

        return geos

    @property
    def data(self):
        geos = self.geos
        return json.dumps({
            'type': 'FeatureCollection',
            'features': geos,
        }, indent=self._indent, cls=NumpyEncoder)


class AddedCalculationNodesGeoJsonSerializer():
    def __init__(self, added_calculationnodes=None, data=None, indent=None):
        if geojson is None:
            raise_import_exception('geojson')
        if added_calculationnodes:
            assert isinstance(added_calculationnodes, AddedCalculationNodes)
        self._data = data
        self._added_calculationnodes = added_calculationnodes
        self._indent = indent

    @property
    def geos(self):
        if self._added_calculationnodes:
            selection = self._added_calculationnodes.to_dict()
        else:
            selection = self._data

        geos = []
        if 'coordinates' not in selection:
            raise ValueError(
                "Can't return data as geojson "
                "for selection without geometries")

        for i in range(selection['id'].shape[-1]):
            pt = geojson.Point(
                [round(x, constants.LONLAT_DIGITS)
                 for x in selection['coordinates'][:, i]])
            cn_meta = [
                    ['object_type', 'added calculation node'],
                    ['nod idx', int(selection['id'][i])],
                    ]
            cn_props = dict(
                object_type='v2_node',
                meta=cn_meta,
                has_1d2d=True,
                node_idx=int(selection['id'][i]),
                size=0,
                )
            feat = geojson.Feature(
                geometry=pt,

                properties=cn_props,
                )
            geos.append(feat)

        return geos

    @property
    def data(self):
        geos = self.geos
        return json.dumps({
            'type': 'FeatureCollection',
            'features': geos,
        }, indent=self._indent, cls=NumpyEncoder)



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function


class DataSource(object):
    """
    Datasource containing a 'dict' like object
    acting as the main datasource for the gridadmin
    models.
    """
    meta = None
    _source = None

    def __init__(self, source, meta=None):
        self.meta = meta
        self._source = source

    def keys(self):
        keys = list(self._source.keys())
        # Append meta to keys if not present
        if 'meta' not in keys:
            keys.append('meta')
        return keys

    def set(self, name, value):
        raise NotImplementedError()

    def get(self, name):
        # meta is special
        if name == 'meta':
            return self.meta
        return self._source.get(name)

    def __getitem__(self, name):
        # meta is special
        if name == 'meta':
            return self.meta
        return self._source[name]



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-

from __future__ import absolute_import
import os
import logging
from collections import OrderedDict
import six
from six.moves import range

try:
    from osgeo import ogr
except ImportError:
    ogr = None

from threedigrid.numpy_utils import reshape_flat_array
from threedigrid.geo_utils import get_spatial_reference
from threedigrid.orm.base.exporters import BaseOgrExporter
from threedigrid.admin import exporter_constants as const

logger = logging.getLogger(__name__)


class LeveeOgrExporter(BaseOgrExporter):
    """
    Exports to ogr formats. You need to set the driver explicitly
    before calling save()
    """
    def __init__(self, levees):
        """
        :param lines: lines.models.Lines instance
        """
        self._levees = levees
        self.supported_drivers = {
            const.GEO_PACKAGE_DRIVER_NAME,
            const.SHP_DRIVER_NAME,
            const.GEOJSON_DRIVER_NAME,
        }
        self.driver = None

    def save(self, file_name, levee_data, target_epsg_code, **kwargs):
        """
        save to file format specified by the driver, e.g. shapefile

        :param file_name: name of the outputfile
        :param line_data: dict of line data
        """
        assert self.driver is not None

        geomtype = 0
        sr = get_spatial_reference(target_epsg_code)

        self.del_datasource(file_name)
        data_source = self.driver.CreateDataSource(file_name)
        layer = data_source.CreateLayer(
            str(os.path.basename(file_name)),
            sr,
            geomtype
        )
        # accounts for 2D only
        fields = OrderedDict([
            ('id', 'int'),
            ('cr_level', 'float'),
            ('mx_depth', 'float'),
        ])

        for field_name, field_type in six.iteritems(fields):
            layer.CreateField(ogr.FieldDefn(
                    field_name, const.OGR_FIELD_TYPE_MAP[field_type])
            )
        _definition = layer.GetLayerDefn()

        for i in range(len(self._levees.geoms)):
            line = ogr.Geometry(ogr.wkbLineString)
            feature = ogr.Feature(_definition)
            linepoints = reshape_flat_array(levee_data['coords'][i]).T
            for x in linepoints:
                line.AddPoint(x[0], x[1])
            feature.SetGeometry(line)
            # for field_name, field_type in fields.iteritems():
            #     raw_value = levee_data[field_name][i]
            #     print("raw_value  ", raw_value)
            #     value = TYPE_FUNC_MAP[field_type](raw_value)
            #     print("value  ", value)

            feature.SetField(str('id'), int(levee_data['id'][i]))
            feature.SetField(
                str('cr_level'), float(levee_data['crest_level'][i])
            )
            feature.SetField(
                str('mx_depth'), float(levee_data['max_breach_depth'][i])
            )

            layer.CreateFeature(feature)
            feature.Destroy()



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

from __future__ import absolute_import
import h5py
import numpy as np
from shapely import wkt
from shapely.geometry import MultiPoint, Point
from threedigrid.admin import constants
from threedigrid.admin.prepare_utils import (
        db_objects_to_numpy_array_dict, add_or_update_datasets)
from six.moves import range
from six.moves import zip


DT_VARIABLE = h5py.special_dtype(vlen=np.dtype('float64'))


def as_numpy_array(array):
    if hasattr(array, 'value'):
        return array.value
    return array


class PrepareLines(object):

    @staticmethod
    def get_1d_object_info(datasource, id_mapper):
        """
        Set content type and content pk's based on id mapping and
        kcu's from calculationcore.

        :return: content_type and content_pk with length lines
        """

        # TODO channel might be different
        LINE_TYPES = [
            constants.TYPE_V2_PIPE, constants.TYPE_V2_CHANNEL,
            constants.TYPE_V2_CULVERT,
            constants.TYPE_V2_ORIFICE, constants.TYPE_V2_WEIR]

        _tmp_kcu = datasource['kcu'].value
        filter_1d = (_tmp_kcu >= 0) & (_tmp_kcu <= 5)

        lik_all = as_numpy_array(datasource['lik'])
        lik_1d = as_numpy_array(datasource['lik'])[filter_1d]

        content_pk = np.zeros(lik_all.shape, dtype='i4')
        content_type = np.zeros(lik_all.shape, dtype='U32')

        _content_pk = np.zeros(lik_1d.shape, dtype='i4')
        _content_type = np.zeros(lik_1d.shape, dtype='U32')

        for line_type in LINE_TYPES:
            line_code = constants.TYPE_CODE_MAP[line_type]
            mapper_idx = id_mapper.obj_slices.get(line_code)
            if mapper_idx is None:
                continue
            seq_ids = id_mapper.id_mapping[mapper_idx]['seq_id']
            for i in range(len(seq_ids)):
                slice1 = np.where(lik_1d == seq_ids[i])
                _content_pk[slice1] = id_mapper.id_mapping[mapper_idx]['pk'][i]
                _content_type[slice1] = line_type

        content_pk[filter_1d] = _content_pk
        content_type[filter_1d] = _content_type

        return content_pk, content_type

    @classmethod
    def make_line_geometries(cls, datasource, threedi_datasource):
        """

        create line geometries based on spatialite geometries but with
        segmentized for calculation core line segments
        :return:
        flattened array of line_geometries with length lines
        array[
            [x1, x2, x3, y1, y2, y3],
            [x1, x2, y1, y2]
        ]
        """

        DB_OBJECTS = [
            threedi_datasource.v2_channels, threedi_datasource.v2_culverts
        ]

        LINE_TYPES = [
            constants.TYPE_V2_CHANNEL, constants.TYPE_V2_CULVERT
        ]

        line_db_dict = dict(list(zip(LINE_TYPES, DB_OBJECTS)))

        size_array = as_numpy_array(datasource['lik']).shape[0]
        line_geometries = np.full(size_array, 0, dtype=DT_VARIABLE)
        start_x = datasource['line_coords'][0][:]
        start_y = datasource['line_coords'][1][:]
        end_x = datasource['line_coords'][2][:]
        end_y = datasource['line_coords'][3][:]
        kcu = datasource['kcu'][:]
        xys = np.array(list(zip(start_x.T, end_x.T, start_y.T, end_y.T)))
        for i in range(len(line_geometries)):
            line_geometries[i] = np.array(xys[i])

        for line_type, db_objects in line_db_dict.items():
            for db_object in db_objects:
                line_idx = np.where(
                    (datasource['content_pk'].value == db_object.pk) &
                    (datasource['content_type'].value == line_type))[0]
                geom = wkt.loads(db_object.the_geom.wkt)
                line_geometries[line_idx] = PrepareLines._cut_geometries(
                    geom,
                    start_x[line_idx], start_y[line_idx],
                    end_x[line_idx], end_y[line_idx],
                    kcu[line_idx])

        return line_geometries

    @staticmethod
    def _cut_geometries(geom, start_x, start_y, end_x, end_y, kcu_array):
        """
        Segmentize line geometry based on start and end points from calc
        line segments

        :param geom: orginal geometry from the sqlite database
        :param start_x: threedicore x-coordinates for start points
        :param start_y: threedicore y-coordinates for start points
        :param end_x: threedicore x-coordinates for end points
        :param end_y: threedicore y-coordinates for end points
        :kcu_array: corresponding kcu values for threedicore coordinates

        :returns  an array of seperate line geometries for number of
        calc lines on geometry, like so::

            array[
                [x1, x2, x3, y1, y2, y3],
                [x1, x2, y1, y2]
            ]
        """

        cut_geometries = np.zeros((len(start_x),), dtype=DT_VARIABLE)
        start_points = MultiPoint(list(zip(start_x, start_y)))
        end_points = MultiPoint(list(zip(end_x, end_y)))

        for piece in range(len(cut_geometries)):
            kcu = kcu_array[piece]
            if kcu == 0:
                start_pnt = geom.interpolate(geom.project(start_points[piece]))
                end_pnt = geom.interpolate(geom.project(end_points[piece]))
            else:
                start_pnt = start_points[piece]
                end_pnt = end_points[piece]
            start_distance = round(geom.project(start_pnt), 3)
            end_distance = round(geom.project(end_pnt), 3)
            coords = list(geom.coords)
            start_set = False
            # no additional calc points
            if start_distance <= 0.0 and end_distance >= geom.length:
                linestring = np.array(
                    [(start_pnt.x, start_pnt.y)] +
                    coords[1:-1] +
                    [(end_pnt.x, end_pnt.y)])
                # "F" means to flatten in column-major (Fortran- style) order
                cut_geometries[piece] = linestring.flatten('F')

            for i, p in enumerate(coords):
                # dist to vertex
                pd = round(geom.project(Point(p)), 3)

                # should not happen but sometimes drawing direction does not
                # correspond with start and endpoints, so flip them
                if start_distance > end_distance:
                    # This is not safe!!!!!!
                    # TODO: check if this actually can happen
                    start_distance, end_distance = end_distance, start_distance
                    start_points, end_points = end_points, start_points

                # check for start point
                if (pd == start_distance) and not start_set:
                    start_pnt = []
                    start_i = i
                    start_set = True
                # should not happen but in case pd (point) and first vertex
                # do not have the same position move pd back
                elif (pd > start_distance) and not start_set:
                    start_pnt = [
                        (start_pnt.x, start_pnt.y)]
                    start_i = i
                    start_set = True
                if pd >= end_distance:
                    linestring = np.array(
                        start_pnt + coords[start_i: i] +
                        [(end_pnt.x, end_pnt.y)])
                    # "F" means to flatten in column-major (Fortran- style)
                    # order
                    cut_geometries[piece] = linestring.flatten('F')
                    break

        return cut_geometries

    @classmethod
    def prepare_datasource(cls, datasource, id_mapper, threedi_datasource,
                           node_coordinates, has_1d):
        if 'id' not in list(datasource.keys()):
            datasource.set(
                'id', np.arange(0, datasource['kcu'].size))

        if has_1d and ('content_pk' not in list(datasource.keys()) or
           'content_type' not in list(datasource.keys())):
            content_pk, content_type =\
                cls.get_1d_object_info(datasource, id_mapper)

            if 'content_pk' not in list(datasource.keys()):
                datasource.set('content_pk', content_pk)
            if 'content_type' not in list(datasource.keys()):
                datasource.set('content_type',
                               [x.encode('ascii') for x in content_type])

        if 'line_coords' not in list(datasource.keys()):
            line = as_numpy_array(datasource['line'])
            x, y = as_numpy_array(node_coordinates[0]),\
                as_numpy_array(node_coordinates[1])
            datasource.set('line_coords', np.array(
                [x[line[0]], y[line[0]], x[line[1]], y[line[1]]]))

        if has_1d and 'line_geometries' not in list(datasource.keys()):
            line_geometries = cls.make_line_geometries(datasource,
                                                       threedi_datasource)
            datasource.set('line_geometries', np.array(line_geometries))


class PrepareChannels(object):

    @staticmethod
    def prepare_datasource(h5py_file, threedi_datasource):
        line_group = h5py_file['lines']

        content_pk = line_group['content_pk'].value
        content_type = line_group['content_type'].value

        channels_field_names = [
            'pk', 'code', 'calculation_type', 'dist_calc_points',
            'connection_node_start_pk', 'connection_node_end_pk',
            'zoom_category'
        ]

        channels_numpy_array_dict = db_objects_to_numpy_array_dict(
            threedi_datasource.v2_channels, channels_field_names)

        add_or_update_datasets(
            line_group, channels_numpy_array_dict,
            channels_field_names,
            channels_numpy_array_dict['pk'], content_pk,
            ignore_mask=content_type != 'v2_channel')


class PreparePipes(object):

    @staticmethod
    def prepare_datasource(h5py_file, threedi_datasource):
        line_group = h5py_file['lines']
        content_pk = line_group['content_pk'].value
        content_type = line_group['content_type'].value

        pipes_field_names = [
            'pk',
            'display_name',
            'invert_level_start_point_raw',
            'invert_level_end_point_raw',
            'friction_type',
            'friction_value',
            'material',
            'sewerage_type',
            'calculation_type',
            'connection_node_start_pk',
            'connection_node_end_pk',
            'zoom_category',
            'cross_section_definition__db_width',
            'cross_section_definition__db_height',
            'cross_section_definition__db_shape']

        pipes_field_name_override = {
            'invert_level_start_point_raw': 'invert_level_start_point',
            'invert_level_end_point_raw': 'invert_level_end_point',
            'cross_section_definition__db_width': 'cross_section_width',
            'cross_section_definition__db_height': 'cross_section_height',
            'cross_section_definition__db_shape': 'cross_section_shape'}

        pipes_numpy_array_dict = db_objects_to_numpy_array_dict(
            threedi_datasource.v2_pipes, pipes_field_names)

        add_or_update_datasets(
            line_group, pipes_numpy_array_dict,
            pipes_field_names,
            pipes_numpy_array_dict['pk'], content_pk,
            ignore_mask=content_type != 'v2_pipe',
            field_name_override=pipes_field_name_override)


class PrepareWeirs(object):

    @staticmethod
    def prepare_datasource(h5py_file, threedi_datasource):
        line_group = h5py_file['lines']
        content_pk = line_group['content_pk'].value
        content_type = line_group['content_type'].value

        weirs_field_names = [
            'pk',
            'code',
            'display_name',
            'discharge_coefficient_negative',
            'discharge_coefficient_positive',
            'sewerage',
            'friction_type',
            'friction_value',
            'crest_type',
            'crest_level',
            'connection_node_start_pk',
            'connection_node_end_pk',
            'zoom_category',
            'cross_section_definition__db_width',
            'cross_section_definition__db_height',
            'cross_section_definition__db_shape']

        weir_field_name_override = {
            'cross_section_definition__db_width': 'cross_section_width',
            'cross_section_definition__db_height': 'cross_section_height',
            'cross_section_definition__db_shape': 'cross_section_shape'}

        weirs_numpy_array_dict = db_objects_to_numpy_array_dict(
            threedi_datasource.v2_weirs, weirs_field_names)

        add_or_update_datasets(
            line_group, weirs_numpy_array_dict,
            weirs_field_names,
            weirs_numpy_array_dict['pk'], content_pk,
            ignore_mask=content_type != 'v2_weir',
            field_name_override=weir_field_name_override
        )


class PrepareOrifices(object):

    @staticmethod
    def prepare_datasource(h5py_file, threedi_datasource):
        line_group = h5py_file['lines']
        content_pk = line_group['content_pk'].value
        content_type = line_group['content_type'].value

        orifices_field_names = [
            'pk',
            'display_name',
            'sewerage',
            'friction_type',
            'friction_value',
            'discharge_coefficient_negative',
            'discharge_coefficient_positive',
            'crest_type',
            'crest_level',
            'connection_node_start_pk',
            'connection_node_end_pk',
            'zoom_category'
        ]

        orifices_numpy_array_dict = db_objects_to_numpy_array_dict(
            threedi_datasource.v2_orifices, orifices_field_names)

        add_or_update_datasets(
            line_group, orifices_numpy_array_dict,
            orifices_field_names,
            orifices_numpy_array_dict['pk'], content_pk,
            ignore_mask=content_type != 'v2_orifice')


class PrepareCulverts(object):

    @staticmethod
    def prepare_datasource(h5py_file, threedi_datasource):
        line_group = h5py_file['lines']
        content_pk = line_group['content_pk'].value
        content_type = line_group['content_type'].value

        culverts_field_names = [
            'pk',
            'code',
            'display_name',
            'discharge_coefficient_negative',
            'discharge_coefficient_positive',
            'friction_type',
            'friction_value',
            'invert_level_start_point_raw',
            'invert_level_end_point_raw',
            'calculation_type',
            'dist_calc_points',
            'connection_node_start_pk',
            'connection_node_end_pk',
            'zoom_category',
            'cross_section_definition__db_width',
            'cross_section_definition__db_height',
            'cross_section_definition__db_shape']

        culverts_field_name_override = {
            'invert_level_start_point_raw': 'invert_level_start_point',
            'invert_level_end_point_raw': 'invert_level_end_point',
            'cross_section_definition__db_width': 'cross_section_width',
            'cross_section_definition__db_height': 'cross_section_height',
            'cross_section_definition__db_shape': 'cross_section_shape'}

        culverts_numpy_array_dict = db_objects_to_numpy_array_dict(
            threedi_datasource.v2_culverts, culverts_field_names)

        add_or_update_datasets(
            line_group, culverts_numpy_array_dict,
            culverts_field_names,
            culverts_numpy_array_dict['pk'], content_pk,
            ignore_mask=content_type != 'v2_culvert',
            field_name_override=culverts_field_name_override)



# -*- coding: utf-8 -*-

from __future__ import absolute_import

# Threedigrid version number is automatic updated with zest.releaser
# the version number in setup.py is updated using the find_version()
__version__ = '1.0.21'



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
"""
Models
++++++

Breaches are better described as possible breach locations
as they represent the intersection between threedicore
1d-2d flow lines and levee geometries from the models spatialite
database. That is, once the kcu (the type classifier, see :ref:`kcu-label`)
of the 1d-2d flow line has the value 56 there actually will be an active
link between the 1d and the 2d calculation nodes.

To query the ``Breaches`` model::

    >>> from threedigrid.admin.breaches.models import Breaches
    >>> from threedigrid.admin.gridadmin import GridH5Admin

    >>> f = 'gridadmin.h5'
    >>> ga = GridH5Admin(f)
    >>> # get all active breaches
    >>> ga.breaches.filter(kcu__eq=56).data

"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import absolute_import
import logging
from threedigrid.admin.breaches import exporters
from threedigrid.admin import constants
from threedigrid.orm.models import Model
from threedigrid.orm.fields import (
    ArrayField, PointArrayField)
from threedigrid.orm.base.fields import IndexArrayField


logger = logging.getLogger(__name__)


class Breaches(Model):
    """
    fields originating from threedicore:

        - levbr (breach width)
        - levmat (material code)
        - levl (exchange level)
        - kcu (type field 55 or 56)

    added fields from spatialite database:

        - content_pk (primary key database)
        - seq_ids (sequence ids generated during input file generation)

    """
    content_pk = ArrayField()
    seq_ids = ArrayField()
    levbr = ArrayField()
    levl = IndexArrayField(to='Lines')
    levmat = ArrayField()
    kcu = ArrayField()

    coordinates = PointArrayField()

    OBJECT_TYPE = constants.TYPE_V2_BREACH

    def __init__(self, *args, **kwargs):
        logger.warning(
            "Deprecation warning: This model is going to be removed "
            " in the near future, "
            "please use breaches under lines (lines.breaches) instead"
        )
        super(Breaches, self).__init__(*args, **kwargs)

        self._exporters = [
            exporters.BreachesOgrExporter(self),
        ]



import json
import logging
from collections import OrderedDict

import numpy as np

from threedigrid.admin import constants
from threedigrid.geo_utils import transform_bbox, raise_import_exception
from threedigrid.orm.base.encoder import NumpyEncoder
from threedigrid.orm.base.models import Model

try:
    import geojson
except ImportError:
    geojson = None


logger = logging.getLogger(__name__)


class GeoJsonSerializer:
    def __init__(self, fields, model=None, indent=None):
        if geojson is None:
            raise_import_exception('geojson')
        self.fields = fields
        if model:
            assert isinstance(model, Model)
        self._model = model
        self._indent = indent

    def save(self, filename, **kwargs):
        with open(filename, "w") as file:
            json.dump(
                {"type": "FeatureCollection", "features": self.geos},
                file,
                indent=self._indent,
                cls=NumpyEncoder,
            )

    @property
    def geos(self):
        geos = []
        if self._model.count == 0:
            return geos
        data = self._model.to_dict()
        content_type = self._model.__contenttype__()
        model_type = type(self._model).__name__
        if content_type == "lines":
            for i in range(data["id"].shape[-1]):
                linepoints = np.round(
                    data['line_geometries'][i].reshape(2, -1)[::-1].T,
                    constants.LONLAT_DIGITS
                )
                line = geojson.LineString(linepoints.tolist())

                properties = fill_properties(
                    self.fields, data, i, model_type
                )
                feat = geojson.Feature(geometry=line, properties=properties)
                geos.append(feat)
        elif content_type in ("nodes", "breaches", "pumps"):
            for i in range(data["id"].shape[-1]):
                coords = np.round(
                    data["coordinates"][:, i], constants.LONLAT_DIGITS
                )
                point = geojson.Point([coords[0], coords[1]])
                properties = fill_properties(
                    self.fields, data, i, model_type
                )
                feat = geojson.Feature(geometry=point, properties=properties)
                geos.append(feat)
        elif content_type == "cells":
            if (self._model.reproject_to_epsg is not None and
                    self._model.reproject_to_epsg != self._model.epsg_code):
                cell_coords = transform_bbox(
                    self._model.reproject_to(
                        self._model.epsg_code
                    ).cell_coords,
                    self._model.epsg_code, self._model.reproject_to_epsg,
                    all_coords=True
                )
            else:
                cell_coords = np.array([
                    data.get('cell_coords')[0], data.get('cell_coords')[3],
                    data.get('cell_coords')[2], data.get('cell_coords')[3],
                    data.get('cell_coords')[2], data.get('cell_coords')[1],
                    data.get('cell_coords')[0], data.get('cell_coords')[1],
                ])
            cell_coords = np.round(cell_coords, constants.LONLAT_DIGITS)
            for i in range(data["id"].shape[-1]):
                left_top = (cell_coords[0][i], cell_coords[1][i])
                right_top = (cell_coords[2][i], cell_coords[3][i])
                right_bottom = (cell_coords[4][i], cell_coords[5][i])
                left_bottom = (cell_coords[6][i], cell_coords[7][i])
                polygon = geojson.Polygon([
                    (left_top, right_top, right_bottom, left_bottom, left_top)
                ])
                properties = fill_properties(
                    self.fields, data, i, model_type
                )
                feat = geojson.Feature(geometry=polygon, properties=properties)
                geos.append(feat)
        elif content_type == "levees":
            for i in range(data["id"].shape[-1]):
                coords = np.round(
                    data["coords"][i].reshape(2, -1), constants.LONLAT_DIGITS
                )
                line = geojson.LineString(zip(coords[1, :], coords[0, :]))
                properties = fill_properties(
                    self.fields, data, i, model_type
                )
                feat = geojson.Feature(geometry=line, properties=properties)
                geos.append(feat)
        else:
            raise ValueError("Unknown content type for %s" % self._model)

        return geos

    @property
    def data(self):
        geos = self.geos
        return json.dumps(
            {"type": "FeatureCollection", "features": geos},
            indent=self._indent,
            cls=NumpyEncoder,
        )


def fill_properties(fields, data, index, model_type=None):
    """Returns a dict containing the keys from `fields` filled from `data`
    at `index`

    :param fields: (list) list of keys (str) to extract from data. The keys
        can can also be a dict, in which case fill_properties will be called
        on its items.
    :param data: (dict) the original data
    :param index: (int) index value to extract from data
    :param model_type: (str, optional) optionally adds the given model_type
    """
    result = OrderedDict()
    for i, field in enumerate(fields):
        if isinstance(field, dict):
            for key, sub_list in field.items():
                result[key] = fill_properties(sub_list, data, index)
        else:
            field_data = data.get(field, None)
            if field_data is not None and field_data.size > 0:
                value = field_data[..., index]
                if value.size == 1:
                    value = value.item()
            else:
                if index == 0:
                    # only log it once
                    logger.warning("missing field %s" % field)
                value = None
            result[field] = value

    if model_type:
        result['model_type'] = model_type

    return result



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

from __future__ import absolute_import
import numpy as np
import logging

logger = logging.getLogger(__name__)

BBOX_LEFT = 0
BBOX_TOP = 1
BBOX_RIGHT = 2
BBOX_BOTTOM = 3


def angle_in_degrees(x0, y0, x1, y1):
    """
    Calculate the angle of lines made up by the points (p1, p2):

        p1[i] = ([x0[i], y0[i]])
        p2[i] = ([x1[i], y1[i]])

    param x0: np.array with x coordinates of point p1[i]
    param y0: np.array with y coordinates of point p1[i]
    param x1: np.array with x coordinates of point p2[i]
    param y1: np.array with y coordinates of point p2[i]

    Returns: the angles of the lines with the (horizontal) x-axis
             in degrees

    Example:
    >>> angle_in_degrees(np.array([0.0]), np.array([0.0]),
    ...                  np.array([1.0]), np.array([1.0]))
    array([ 45.])
    """

    assert isinstance(x0, np.ndarray)
    assert isinstance(y0, np.ndarray)
    assert isinstance(x1, np.ndarray)
    assert isinstance(y1, np.ndarray)

    # Calculate the differences
    dx, dy = x1 - x0, y1 - y0

    # Divide by zero (dx) solution:
    #   skip (dy/dx)[i] where dx[i] == 0 and replace
    #   the (dy/dx)[i] value with np.inf when dy >= 0 and
    #   -np.inf when dy < 0

    # Create an array with the size of dy with np.inf where
    # dy >=0 and -np.inf where dy < 0
    out_inf = np.full_like(dy, np.inf)
    out_inf[dy < 0] = -np.inf

    # compute dy/dx, but use the out_inf[i] value if dx[i]==0
    dy_divided_by_dx_with_inf = np.divide(
        dy, dx, out=out_inf, where=dx != 0)

    # Return the values of the inverse-tan in degrees
    # np.inf gives 90 and -np.inf gives -90
    return np.degrees(np.arctan(dy_divided_by_dx_with_inf))


def lines_to_bbox_array(lines):
    """
    Convert lines:
        x1_array=lines[0], y1_array=lines[1],
        x2_array=lines[2], y2_array=lines[3]

    to bbox_array:
        bbox[i] = [left, top, right, bottom]

    :param lines:
        x1_array=lines[0], y1_array=lines[1],
        x2_array=lines[2], y2_array=lines[3]

    :return: np.array with bboxes
    """
    # Use min/max, don't assume the x1[i] to always be left,
    # y1[i] to be top etc..
    return np.stack(
        (np.fmin(lines[0], lines[2]), np.fmin(lines[1], lines[3]),
         np.fmax(lines[0], lines[2]), np.fmax(lines[1], lines[3])), axis=1)


def select_lines_by_bbox(lines, bbox, include_intersections=False):
    """
    Return a boolean mask array for the lines that are in bbox

    :param lines: np.array
                x1_array=lines[0], y1_array=lines[1],
                x2_array=lines[2], y2_array=lines[3]

                line[i] = (x1[i], y1[i]) - (x2[i], y2[i])

    :param bbox: np.array [left, top, right, bottom]
    :return: a boolean mask array with 'True' for the
             points that are in the bbox

    """
    assert isinstance(lines, np.ndarray)

    # We can treat lines same like bboxes
    # if we transform the lines array's to bboxes
    bbox_array = lines_to_bbox_array(lines)

    if include_intersections:
        return get_bbox_intersect_bool_mask(bbox_array, bbox)

    return get_bbox_in_bool_mask(bbox_array, bbox)


def get_bbox_in_bool_mask(bbox_array, bbox):
    """
    Return a boolean mask with 'True' when bbox_array[i] is in bbox2

    :param bbox_array: np.array with bboxes [left, top, right, bottom]
    :param bbox: np.array to match with bbox_array items
                 [left, top, right, bottom]
    :return: A boolean mask with 'True' when bbox_array[i] is in
             with bbox

    Example:
    >>> get_bbox_in_bool_mask(
    ...     bbox_array=np.array(
    ...         [[5, 5, 10, 10], [15, 15, 30, 30], [50, 50, 60, 60]]),
    ...     bbox=np.array([0, 0, 20, 20]))
    array([ True, False, False], dtype=bool)
    """

    assert isinstance(bbox_array, np.ndarray)

    return ((bbox[BBOX_LEFT] <= bbox_array[:, BBOX_LEFT]) &
            (bbox[BBOX_TOP] <= bbox_array[:, BBOX_TOP]) &
            (bbox[BBOX_RIGHT] >= bbox_array[:, BBOX_RIGHT]) &
            (bbox[BBOX_BOTTOM] >= bbox_array[:, BBOX_BOTTOM])).flatten()


def get_bbox_by_point(pnt, values):
    """

    :param pnt: list or tuple of xy values
    :param values: arrays of xy's of the lower left and
        upper right corner, e.g. [[x,...], [y,...], [x1,...], [y1,...]]]
    :returns the bbox (lower left and upper right corner)
        in values which contains pnt
    Example:
    >>> get_bbox_by_point(
    ...     pnt=(1,5),
    ...     values=np.array([[0,2], [0,2], [6,6], [10,11]]))
    array([ True, False], dtype=bool)
    """
    if isinstance(pnt, (list, tuple)):
        pnt = np.array(pnt)

    assert isinstance(pnt, np.ndarray)
    # pnt --> np.array(x, y)
    lleft = np.array((values[0], values[1]))
    uright = np.array((values[2], values[3]))

    pre_sel_low = pnt >= lleft.T
    pre_sel_up = pnt <= uright.T
    return np.all((pre_sel_low & pre_sel_up), axis=1)


def get_bbox_intersect_bool_mask(bbox_array, bbox):
    """
    Return a boolean mask with 'True' when bbox_array[i] intersects with bbox2

    :param bbox_array: np.array with bboxes [left, top, right, bottom]
    :param bbox: np.array to match with bbox_array items
                 [left, top, right, bottom]
    :return: A boolean mask with 'True' when bbox_array[i] intersects
             with bbox

    Example:
    >>> get_bbox_intersect_bool_mask(
    ...     bbox_array=np.array(
    ...         [[5, 5, 10, 10], [15, 15, 30, 30], [50, 50, 60, 60]]),
    ...     bbox=np.array([0, 0, 20, 20]))
    array([ True,  True, False], dtype=bool)
    """

    assert isinstance(bbox_array, np.ndarray)

    # Check which ones do not intersect with:
    #    bbox[BBOX_LEFT] > bbox_array[:, BBOX_RIGHT] etc..
    #
    #              bbox_array[i, BBOX_RIGHT]
    #              V
    #          ----- -----
    #          |   | |   |
    #          ----- -----
    #                ^
    #                bbox[BBOX_LEFT]
    #
    # Invert the boolean array and flatten.
    return np.invert(
            (bbox[BBOX_LEFT] > bbox_array[:, BBOX_RIGHT]) |
            (bbox[BBOX_RIGHT] < bbox_array[:, BBOX_LEFT]) |
            (bbox[BBOX_TOP] > bbox_array[:, BBOX_BOTTOM]) |
            (bbox[BBOX_BOTTOM] < bbox_array[:, BBOX_TOP])).flatten()


def get_bbox_intersections(bbox_array, bbox):
    """
    Return the intersections and indices of bbox_array that intersect
    with bbox2

    :param bbox_array: np.array with bboxes [left, top, right, bottom]
    :param bbox: np.array to match with bbox_array items
                 [left, top, right, bottom]
    :return: the intersections and indices of bbox_array which intersect with
             bbox2

    Example:
    >>> get_bbox_intersections(
    ...     bbox_array=np.array(
    ...         [[5, 5, 10, 10], [15, 15, 30, 30], [50, 50, 60, 60]]),
    ...     bbox=np.array([0, 0, 20, 20]))
    (array([[ 5,  5, 10, 10],
           [15, 15, 20, 20]]), array([0, 1]))
    """

    assert isinstance(bbox_array, np.ndarray)

    intersect_mask = get_bbox_intersect_bool_mask(bbox_array, bbox)
    nodes_intersect = bbox_array[intersect_mask]

    # Calculate the intersections by getting the:
    #     max values for: left, top
    #     miniminum for: right, bottom
    #
    #     For example (left):
    #
    #          bbox_array[intersection_mask][i, BBOX_LEFT]
    #          V
    #          -----
    #          |   |
    #          -----
    #             -----
    #             |   |
    #             -----
    #             ^
    #             bbox[BBOX_LEFT]
    #
    #      intersection left value is max of both values.
    intersections = np.concatenate(
        (np.fmax(nodes_intersect[:, 0:2], bbox[0:2]),
         np.fmin(nodes_intersect[:, 2:4], bbox[2:4])), axis=1)
    return (intersections, np.argwhere(intersect_mask).flatten())


def reshape_flat_array(flat_array):
    return flat_array.reshape(2, -1)


def create_np_lookup_index_for(search_array, index_array):
    """
    Return a np.ndarray which can be used to lookup the indices
    from index_array

    This can be used to map values of two array's like:

    search_array = np.array([7, 8, 2])
    index_array = np.array([2, 7, 8])
    lookup = create_np_lookup_index_for(search_array, index_array)
    indices_of_search_array_in_index_array = lookup[search_array]

    index_array[indices_of_search_array_in_index_array] == search_array
    (If all elements could be found)

    :param search_array: the np.ndarray with values that should be available
                         for the lookup
    :param index_array:  the np.ndarray from which the indexes should be
                         searchable by value

    :return: an np.ndarray with the values of index_array as indices, or -1
             if the value is not in index_array

    examples:

    >>> create_np_lookup_index_for(np.array([7, 8, 2]), np.array([2, 7, 8]))
    array([1, 2, 0])
    """

    sort_idx = np.argsort(index_array)
    lookup = sort_idx[np.searchsorted(
        index_array, search_array, sorter=sort_idx)]
    return lookup


def get_smallest_uint_dtype(maxval):
    """Returns smallest unsigned integer datatype for holding maxval."""
    if maxval < 0:
        raise ValueError("Value cannot be negative")
    dtypes = [np.uint8, np.uint16, np.uint32, np.uint64]
    for dt in dtypes:
        if maxval <= np.iinfo(dt).max:
            return dt
    raise ValueError(
        "Value of %s exceeds all possible maximum dtype values." % maxval)



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

from __future__ import absolute_import
import numpy as np

from .base.fields import ArrayField
from threedigrid.numpy_utils import angle_in_degrees
from threedigrid.numpy_utils import get_bbox_by_point
from threedigrid.numpy_utils import reshape_flat_array
from threedigrid.numpy_utils import select_lines_by_bbox
from threedigrid.geo_utils import select_points_by_tile, \
    select_geoms_by_geometry, raise_import_exception
from threedigrid.geo_utils import select_lines_by_tile
from threedigrid.geo_utils import select_points_by_bbox
from threedigrid.geo_utils import transform_xys
from six.moves import map

try:
    from pyproj import Transformer
except ImportError:
    Transformer = None

try:
    import shapely
    from shapely.geometry import Polygon, Point, asLineString, asPolygon
except ImportError:
    shapely = None


class GeomArrayField(ArrayField):
    """
    Base geometry field
    """
    def reproject(self, values, source_epsg, target_epsg):
        raise NotImplementedError()

    def to_centroid(self, values):
        raise NotImplementedError()

    def get_mask_by_geometry(self, geometry, values):
        geoms = self._to_shapely_geom(values)

        selected_geoms = select_geoms_by_geometry(geoms, geometry)

        mask = np.zeros((len(values.T),), dtype=bool)
        for geom in selected_geoms:
            mask[geom.index] = True

        return mask

    def _to_shapely_geom(self, values):
        """Returns a list of shapely geometries created from values

        :param values: coordinates
        :return: list of shapely geometries
        """
        raise NotImplementedError()


class PointArrayField(GeomArrayField):
    def reproject(self, values, source_epsg, target_epsg):
        """
        Reproject the point coordinates: x_array=values[0], y_array=values[1]
        from source_epsg to target_epsg.
        """
        return transform_xys(
            values[0], values[1],
            source_epsg, target_epsg)

    def get_mask_by_bbox(self, bbox, values, include_intersections=False):
        """
        Return as boolean mask (np.array) for line/bbox in "values"
            x_array=values[0], y_array=values[1]
        within bbox (x1, y1, x2, y2)
        """
        # For points include_intersections can be ignored
        return select_points_by_bbox(values, bbox)

    def get_mask_by_tile(self, tile_xyz, target_epsg, values,
                         include_intersections=False):
        """
        Return as boolean mask (np.array) for points in "values"
        within tile_xyz bbox projected to target_epsg.
        """
        # For points include_intersections can be ignored
        return select_points_by_tile(
            tile_xyz, target_epsg, values)

    def to_centroid(self, values):
        """
        Returns: the centroid of the point coordinates:
                 x_array=values[0], y_array=values[1]
        """
        return values

    def _to_shapely_geom(self, values):
        if shapely is None:
            raise_import_exception('shapely')

        points = []
        for i, coord in enumerate(values.T):
            point = Point(coord[0], coord[1])
            point.index = i  # the index is used in get_mask_by_geometry
            points.append(point)
        return points


class LineArrayField(GeomArrayField):
    """
    Field which handles line/bbox geoms values:

        x1_array=values[0], y1_array=values[1],
        x2_array=values[2], y2_array=values[3]
    """

    def reproject(self, values, source_epsg, target_epsg):
        """
        Reproject the line/bbox coordinates:
            x1_array=values[0], y1_array=values[1],
            x2_array=values[2], y2_array=values[3]
        from source_epsg to target_epsg.
        """
        return np.vstack((
                transform_xys(
                    values[0], values[1],
                    source_epsg, target_epsg),
                transform_xys(
                    values[2], values[3],
                    source_epsg, target_epsg)))

    def get_mask_by_point(self, pnt, values):
        return get_bbox_by_point(pnt, values)

    def get_mask_by_bbox(self, bbox, values,
                         include_intersections=False):
        """
        Return as boolean mask (np.array) for line/bbox in "values"
            x1_array=values[0], y1_array=values[1],
            x2_array=values[2], y2_array=values[3]
        within bbox (x1, y1, x2, y2)
        """
        return select_lines_by_bbox(values, bbox, include_intersections)

    def get_mask_by_tile(self, tile_xyz, target_epsg, values,
                         include_intersections=False):
        """
        Return as boolean mask (np.array) for line/bbox in "values"
            x1_array=values[0], y1_array=values[1],
            x2_array=values[2], y2_array=values[3]
        within tile_xyz bbox projected to target_epsg.
        """
        return select_lines_by_tile(
            tile_xyz, target_epsg, values, include_intersections)

    def to_centroid(self, values):
        """
        :return: the centroid (float) for the line coordinates:
                    x1_array=values[0], y1_array=values[1],
                    x2_array=values[2], y2_array=values[3]
        """
        return np.array(
            ((values[0] + values[2]) / 2.0,
             (values[1] + values[3]) / 2.0))

    def get_angles_in_degrees(self, values):
        """
        Returns: the angles in degrees of the lines
        made up by the points:

            x1_array=values[0], y1_array=values[1],
            x2_array=values[2], y2_array=values[3]

        with the (horizontal) x-axis
        """
        return angle_in_degrees(
                values[0], values[1], values[2], values[3])

    def _to_shapely_geom(self, values):
        if shapely is None:
            raise_import_exception('shapely')

        lines = []
        for i, coords in enumerate(values.T):
            line = asLineString(coords.reshape((2, -1)))
            line.index = i  # the index is used in get_mask_by_geometry
            lines.append(line)
        return lines


class MultiLineArrayField(GeomArrayField):

    def reproject(self, values, source_epsg, target_epsg):
        """
        Reproject the line/bbox coordinates:
            x1_array=values[0], y1_array=values[1],
            x2_array=values[2], y2_array=values[3]
        from source_epsg to target_epsg.
        """
        reshaped_values = list(map(reshape_flat_array, values))

        # Pyproj has (a lot) more overhead
        # from 2.0.1 for reprojecting.
        # so created the Transform once
        # and reuse it.
        if Transformer is not None:
            # Only use the Transformer if it is present
            # in Python 3 it works well, however in Python 2.7
            # it seams to break
            transformer = Transformer.from_proj(
                int(source_epsg), int(target_epsg))

            transform_values = [
                np.array(transformer.transform(x[0], x[1])).flatten()
                for x in reshaped_values
            ]
        else:
            transform_values = [
                transform_xys(x[0], x[1], source_epsg, target_epsg).flatten()
                for x in reshaped_values
            ]

        return np.array(transform_values)

    def _to_shapely_geom(self, values):
        if shapely is None:
            raise_import_exception('shapely')

        multilines = []
        for i, coords in enumerate(values):
            line = asLineString(coords.reshape((2, -1)).T)
            line.index = i  # the index is used in get_mask_by_geometry
            multilines.append(line)
        return multilines


class PolygonArrayField(GeomArrayField):
    """
    Field which handles line/bbox geoms values:

        x1_array=values[0], y1_array=values[1],
        x2_array=values[2], y2_array=values[3]

    """
    def get_mask_by_point(self, pnt, values):

        return get_bbox_by_point(pnt, values)

    def reproject(self, values, source_epsg, target_epsg):
        """
        Reproject the cell coordinates (lower left and
        upper right corners):
            x1_array=values[0], y1_array=values[1],
            x2_array=values[2], y2_array=values[3],
        from source_epsg to target_epsg.
        """
        return np.vstack((
                np.array(transform_xys(
                    values[0], values[1],
                    source_epsg, target_epsg)
                ),
                np.array(transform_xys(
                    values[2], values[3],
                    source_epsg, target_epsg)
                ),
        ))

    def _to_shapely_geom(self, values):
        if shapely is None:
            raise_import_exception('shapely')

        polygons = []
        for i, coords in enumerate(values):
            polygon = asPolygon(coords.reshape((2, -1)).T)
            polygon.index = i  # the index is used in get_mask_by_geometry
            polygons.append(polygon)
        return polygons


class BboxArrayField(LineArrayField):

    def _to_shapely_geom(self, values):
        if shapely is None:
            raise_import_exception('shapely')

        polygons = []
        for i, coord in enumerate(values.T):
            # convert the bbox bottom-left and upper-right into a polygon
            polygon = Polygon(
                [
                    (coord[0], coord[1]),
                    (coord[0], coord[3]),
                    (coord[2], coord[3]),
                    (coord[2], coord[1])
                ]
            )
            polygon.index = i  # the index is used in get_mask_by_geometry
            polygons.append(polygon)
        return polygons



# -*- coding: utf-8 -*-
"""help text functions for threedigrid management commands"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from collections import OrderedDict

from threedigrid.admin.gridadmin import GridH5Admin
import six


def model_overview(grid_file):
    """

    :param grid_file: path to hdf5 grid file
    :returns formatted string like so::

        threedicore version:     0-20180219-587b996-1
        has 1d:                  True
        has 2d:                  True
        has groundwater:         False
        has levees:              True
    """

    grid = GridH5Admin(grid_file)
    _attrs = OrderedDict([
        ('model_slug', ''),
        ('threedicore_version', ''),
        ('threedi_version', ''),
        ('has_1d', ''),
        ('has_2d', ''),
        ('has_groundwater', ''),
        ('has_levees', ''),
        ('has_breaches', ''),
        ('has_pumpstations', ''),
    ])

    for k in six.iterkeys(_attrs):
        if not hasattr(grid, k):
            continue
        attr = getattr(grid, k)
        if attr is None:
            continue
        _attrs[k] = '{:24} {}'.format(k.replace('_', ' ') + ':', attr)

    return '\n'.join((v for v in _attrs.values() if v != '')) + '\n'



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

from __future__ import absolute_import
from threedigrid.orm.models import Model
from threedigrid.orm.fields import (
    ArrayField, LineArrayField, PointArrayField)


class Pumps(Model):
    display_name = ArrayField()
    content_pk = ArrayField()
    type = ArrayField()
    node1_id = ArrayField()
    node2_id = ArrayField()
    bottom_level = ArrayField()
    start_level = ArrayField()
    lower_stop_level = ArrayField()
    capacity = ArrayField()
    coordinates = PointArrayField()
    # coordinates is the centroid of
    # node_coordinates if both set, else
    # the one that is set.
    node_coordinates = LineArrayField()
    # [ [node1_x], [node1_y], [node2_x], [node2_y]]
    # Note: -9999 if nodeX_id is -9999 (not set)
    zoom_category = ArrayField()



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

from __future__ import absolute_import
from threedigrid.orm.base.timeseries_mixin import ResultMixin
from threedigrid.orm.base.options import ModelMeta
import six


class BreachesResultsMixin(ResultMixin):

    class Meta(six.with_metaclass(ModelMeta)):
        field_attrs = ['units', 'long_name', 'standard_name']

        # customize field names
        composite_fields = {
            'breach_depth': ['Mesh1D_breach_depth'],
            'breach_width': ['Mesh1D_breach_width']}

    def __init__(self, **kwargs):
        """Instantiate a breach with netcdf results.

        Variables stored in the netcdf and related to breaches are dynamically
        added as attributes as TimeSeriesArrayField.

        :param kwargs:
        """
        super(BreachesResultsMixin, self).__init__(**kwargs)


class BreachesAggregateResultsMixin(ResultMixin):

    class Meta(six.with_metaclass(ModelMeta)):
        field_attrs = ['units', 'long_name']

        base_composition = {
            'breach_depth': ['Mesh1D_breach_depth'],
            'breach_width': ['Mesh1D_breach_width']
        }
        composition_vars = {
            'breach_depth': ['avg', 'min', 'max'],
            'breach_width': ['avg', 'min', 'max'],
        }

    def __init__(self, **kwargs):
        """Instantiate a line with netcdf results.

        Variables stored in the netcdf and related to lines are dynamically
        added as attributes as TimeSeriesArrayField.

        :param netcdf_keys: list of netcdf variables
        :param kwargs:
        """
        super(BreachesAggregateResultsMixin, self).__init__(**kwargs)
